@@ -309,16 +309,17 @@ class SniEndPoint(EndPoint):
 
 class SniEndPointFactory(EndPointFactory):
 
-    def __init__(self, proxy_address, port):
+    def __init__(self, proxy_address, port, node_domain=None):
         self._proxy_address = proxy_address
         self._port = port
+        self._node_domain = node_domain
 
     def create(self, row):
         host_id = row.get("host_id")
         if host_id is None:
             raise ValueError("No host_id to create the SniEndPoint")
-
-        return SniEndPoint(self._proxy_address, str(host_id), self._port)
+        address = "{}.{}".format(host_id, self._node_domain) if self._node_domain else str(host_id)
+        return SniEndPoint(self._proxy_address, str(address), self._port)
 
     def create_from_sni(self, sni):
         return SniEndPoint(self._proxy_address, sni, self._port)
@@ -769,6 +770,8 @@ class Connection(object):
 
     _is_checksumming_enabled = False
 
+    _on_orphaned_stream_released = None
+
     @property
     def _iobuf(self):
         # backward compatibility, to avoid any change in the reactors
@@ -778,8 +781,8 @@ class Connection(object):
                  ssl_options=None, sockopts=None, compression=True,
                  cql_version=None, protocol_version=ProtocolVersion.MAX_SUPPORTED, is_control_connection=False,
                  user_type_map=None, connect_timeout=None, allow_beta_protocol_version=False, no_compact=False,
-                 ssl_context=None, owning_pool=None, shard_id=None, total_shards=None):
-
+                 ssl_context=None, owning_pool=None, shard_id=None, total_shards=None,
+                 on_orphaned_stream_released=None):
         # TODO next major rename host to endpoint and remove port kwarg.
         self.endpoint = host if isinstance(host, EndPoint) else DefaultEndPoint(host, port)
 
@@ -801,7 +804,7 @@ class Connection(object):
         self._continuous_paging_sessions = {}
         self._socket_writable = True
         self.orphaned_request_ids = set()
-        self._owning_pool = owning_pool
+        self._on_orphaned_stream_released = on_orphaned_stream_released
 
         if ssl_options:
             self._check_hostname = bool(self.ssl_options.pop('check_hostname', False))
@@ -900,7 +903,7 @@ class Connection(object):
                     break
                 except Exception as ex:
                     log.debug("port=%d couldn't bind cause: %s", port, str(ex))
-            log.debug(f'connection (%r) port=%d should be shard_id=%d', id(self), port, port % self.total_shards)
+            log.debug('connection (%r) port=%d should be shard_id=%d', id(self), port, port % self.total_shards)
 
         self._socket.connect(sockaddr)
 
@@ -1247,8 +1250,8 @@ class Connection(object):
                         self.in_flight -= 1
                         self.orphaned_request_ids.remove(stream_id)
                         need_notify_of_release = True
-                if need_notify_of_release and self._owning_pool:
-                    self._owning_pool.on_orphaned_stream_released()
+                if need_notify_of_release and self._on_orphaned_stream_released:
+                    self._on_orphaned_stream_released()
 
                 try:
                     callback, decoder, result_metadata = self._requests.pop(stream_id)
